"""Constants for pyskyqremote- GB."""

SCHEDULE_URL = "http://awk.epgsky.com/hawk/linear/schedule/{1}/{0}"
LIVE_IMAGE_URL = "https://images.metadata.sky.com/pd-image/{0}/16-9"
PVR_IMAGE_URL = "https://images.metadata.sky.com/pd-image/{0}/16-9"
CHANNEL_IMAGE_URL = "https://d2n0069hmnqmmx.cloudfront.net/epgdata/1.0/newchanlogos/600/600/skychb{0}.png"  # also at https://epgstatic.sky.com/...
